@@ -17,11 +17,17 @@
 ** implied warranty.
 */
 
-#define VERSION "2.37.3 (13 July 1999) +netpbm"
-
-/* BJH 20000303:  fix include statement so dependencies work out right.
+/* 
+   BJH 20000408:  remove BIGGRAYS stuff -- it's not an option anymore.
+                  MAXMAXVAL renamed to OVERALLMAXVAL.
+   BJH 20000303:  fix include statement so dependencies work out right.
 */
 
+/* GRR 19991203:  incorporated fix by Rafal Rzeczkowski <rzeczkrg@mcmaster.ca>:
+ *  gray images with exactly 16 shades were being to 8-bit grayscale rather
+ *  than 4-bit palette due to misuse of the pm_maxvaltobits() function.  Also
+ *  moved VERSION to new version.h header file. */
+
 /* GRR 19990713:  fixed redundant freeing of png_ptr and info_ptr in setjmp()
  *  blocks and added png_destroy_write_struct() and pm_close(ifp) in each
  *  pm_error() block.  */
@@ -48,6 +54,7 @@
 #define GRR_GRAY_PALETTE_FIX
 
 #include <png.h>	/* includes zlib.h and setjmp.h */
+#define VERSION "2.37.4 (5 December 1999) +netpbm"
 #include "pnm.h"
 
 #include "ppmcmap.h"
@@ -56,6 +63,14 @@ typedef struct _jmpbuf_wrapper {
   jmp_buf jmpbuf;
 } jmpbuf_wrapper;
 
+/* GRR 19991205:  this is used as a test for pre-1999 versions of netpbm and
+ *   pbmplus vs. 1999 or later (in which pm_close was split into two) 
+ */
+#ifdef PBMPLUS_RAWBITS
+#  define pm_closer pm_close
+#  define pm_closew pm_close
+#endif
+
 #ifndef TRUE
 #  define TRUE 1
 #endif
@@ -257,8 +272,7 @@ FILE *tfp;
   xel p;	/*        (pnm.h)                  (ppm.h)  */
   int rows, cols, format;
   xelval maxval;
-  long int scaleval;   
-    /* bjh changed type of scaleval to avoid compiler warnings 00.03.02 */
+  xelval scaleval;
   xelval value;
   pixel transcolor;
   int mayscale;
@@ -315,20 +329,20 @@ FILE *tfp;
   png_ptr = png_create_write_struct (PNG_LIBPNG_VER_STRING,
     &pnmtopng_jmpbuf_struct, pnmtopng_error_handler, NULL);
   if (png_ptr == NULL) {
-    pm_close (ifp);
+    pm_closer (ifp);
     pm_error ("cannot allocate LIBPNG structure");
   }
 
   info_ptr = png_create_info_struct (png_ptr);
   if (info_ptr == NULL) {
     png_destroy_write_struct (&png_ptr, (png_infopp)NULL);
-    pm_close (ifp);
+    pm_closer (ifp);
     pm_error ("cannot allocate LIBPNG structures");
   }
 
   if (setjmp (pnmtopng_jmpbuf_struct.jmpbuf)) {
     png_destroy_write_struct (&png_ptr, &info_ptr);
-    pm_close (ifp);
+    pm_closer (ifp);
     pm_error ("setjmp returns error condition (1)");
   }
 
@@ -345,9 +359,9 @@ FILE *tfp;
   }
 
   if (pnm_type == PGM_TYPE)
-    maxmaxval = PGM_MAXMAXVAL;
+    maxmaxval = PGM_OVERALLMAXVAL;
   else if (pnm_type == PPM_TYPE)
-    maxmaxval = PPM_MAXMAXVAL;
+    maxmaxval = PPM_OVERALLMAXVAL;
 
   if (transparent > 0)   /* -1 or 1 are the only possibilities so far */
     transcolor = ppm_parsecolor (transstring, maxmaxval);
@@ -356,7 +370,7 @@ FILE *tfp;
     alpha_mask = pgm_readpgm (afp, &alpha_cols, &alpha_rows, &alpha_maxval);
     if (alpha_cols != cols || alpha_rows != rows) {
       png_destroy_write_struct (&png_ptr, &info_ptr);
-      pm_close (ifp);
+      pm_closer (ifp);
       pm_error ("dimensions for image and alpha mask do not agree");
     }
     /* check if the alpha mask can be represented by a single transparency
@@ -433,7 +447,7 @@ FILE *tfp;
   if (pnm_type != PBM_TYPE || alpha) {
     if (maxval > 65535 && !downscale) {
       png_destroy_write_struct (&png_ptr, &info_ptr);
-      pm_close (ifp);
+      pm_closer (ifp);
       pm_error ("can only handle files up to 16-bit (use -downscale to override");
     }
 
@@ -458,9 +472,8 @@ FILE *tfp;
       } else {
         /* to do:  this can fail; xels aren't big enough - larger pixels &
          * maxvals are needed */
-        if (sizeof(maxval) < 2) {   /* unsigned char by default */
-          pm_message ("cannot rescale to 16-bit; "
-                      "rescaling to 8-bit instead (maxval = %d)",
+        if (sizeof(scaleval) < 2) {   /* unsigned char by default */
+          pm_message ("cannot rescale to 16-bit; rescaling to 8-bit instead (maxval = %d)",
             maxval);
           scaleval = 255;
         } else {
@@ -647,7 +660,7 @@ FILE *tfp;
        - for ppm files if we have <= 256 colors
        - for alpha if we have <= 256 color/transparency pairs
        - for pgm files (with or without alpha) if the number of bits needed for
-         the gray-transparency pairs is smaller then the number of bits needed
+         the gray-transparency pairs is smaller than the number of bits needed
          for maxval
      When maxval > 255, we never write a paletted image.
   */
@@ -670,8 +683,9 @@ FILE *tfp;
     if (verbose) {
       pm_message ("%d colors found", colors);
     }
-    if ((chv == (colorhist_vector) NULL) ||    /* GRR:    v--- was `=' here */
-        (pnm_type == PGM_TYPE && pm_maxvaltobits(colors) > 
+    /* RR:  note that if colors == 16, maxval == 15 (i.e., range == 0-15) */
+    if ((chv == (colorhist_vector) NULL) ||    /* GRR:      v-- was `=' here */
+        (pnm_type == PGM_TYPE && pm_maxvaltobits(colors-1) > 
                                 (pm_maxvaltobits(maxval) / 2)) ||
         (pnm_type == PPM_TYPE && maxval > 255)) {
       chv = NULL;
@@ -692,7 +706,7 @@ FILE *tfp;
               == NULL)
           {
             png_destroy_write_struct (&png_ptr, &info_ptr);
-            pm_close (ifp);
+            pm_closer (ifp);
             pm_error ("out of memory allocating alpha/palette entries");
           }
           alphas_of_color_cnt[i] = 0;
@@ -807,7 +821,7 @@ FILE *tfp;
       depth = 1;
     else {
       png_destroy_write_struct (&png_ptr, &info_ptr);
-      pm_close (ifp);
+      pm_closer (ifp);
       pm_error (" (can't happen) undefined maxval");
     }
 
@@ -835,7 +849,7 @@ FILE *tfp;
 
   if (setjmp (pnmtopng_jmpbuf_struct.jmpbuf)) {
     png_destroy_write_struct (&png_ptr, &info_ptr);
-    pm_close (ifp);
+    pm_closer (ifp);
     pm_error ("setjmp returns error condition (2)");
   }
 
@@ -932,7 +946,7 @@ FILE *tfp;
       /* indices should have just crossed paths */
       if (bot_idx != top_idx + 1) {
         png_destroy_write_struct (&png_ptr, &info_ptr);
-        pm_close (ifp);
+        pm_closer (ifp);
         pm_error ("internal inconsistency: remapped bot_idx = %d, top_idx = %d",
           bot_idx, top_idx);
       }
@@ -987,7 +1001,7 @@ FILE *tfp;
     } else {
       if (transparent > 0) {
         png_destroy_write_struct (&png_ptr, &info_ptr);
-        pm_close (ifp);
+        pm_closer (ifp);
         pm_error (" (can't happen) transparency AND alpha");
       }
     }
@@ -1085,7 +1099,7 @@ FILE *tfp;
   if ((line = (png_byte *) malloc (cols*8)) == NULL)
   {
     png_destroy_write_struct (&png_ptr, &info_ptr);
-    pm_close (ifp);
+    pm_closer (ifp);
     pm_error ("out of memory allocating PNG row buffer");
   }
 
@@ -1125,7 +1139,7 @@ FILE *tfp;
           *pp++ = PPM_GETB (p)&0xff;
         } else {
           png_destroy_write_struct (&png_ptr, &info_ptr);
-          pm_close (ifp);
+          pm_closer (ifp);
           pm_error (" (can't happen) undefined color_type");
         }
         if (info_ptr->color_type & PNG_COLOR_MASK_ALPHA) {
@@ -1138,12 +1152,28 @@ FILE *tfp;
     }
   }
   png_write_end (png_ptr, info_ptr);
+
+
+#if 0
+  /* The following code may be intended to solve some segfault problem
+     that arises with png_destroy_write_struct().  The latter is the
+     method recommended in the libpng documentation and this program 
+     will not compile under Cygwin because the Windows DLL for libpng
+     does not contain png_write_destroy() at all.  Since the author's
+     comment below does not make it clear what the segfault issue is,
+     we cannot consider it.  -Bryan 00.09.15
+*/
+
   png_write_destroy (png_ptr);
   /* flush first because free(png_ptr) can segfault due to jmpbuf problems
      in png_write_destroy */
   fflush (stdout);
   free (png_ptr);
   free (info_ptr);
+#else
+  png_destroy_write_struct(&png_ptr, &info_ptr);
+#endif
+
   for (i = 0 ; i < num_alphas_of_color ; i++)
     free(alphas_of_color[i]);
 }
@@ -1321,14 +1351,9 @@ char *argv[];
         PNG_LIBPNG_VER_STRING, png_libpng_ver);
       fprintf(stderr, "   Compiled with zlib %s; using zlib %s.\n",
         ZLIB_VERSION, zlib_version);
-#ifdef PGM_BIGGRAYS
-      fprintf(stderr,
-        "   Compiled with 16-bit netpbm support (PGM_BIGGRAYS defined).\n");
-#else
-      fprintf(stderr,
-        "   Compiled with %d-bit netpbm support (PPM_MAXMAXVAL = %d).\n",
-        pm_maxvaltobits (PPM_MAXMAXVAL), PPM_MAXMAXVAL);
-#endif
+      fprintf(stderr,    
+        "   Compiled with %d-bit netpbm support (PPM_OVERALLMAXVAL = %d).\n",
+        pm_maxvaltobits (PPM_OVERALLMAXVAL), PPM_OVERALLMAXVAL);
       fprintf(stderr, "\n");
       pm_usage (usage);
     }
@@ -1357,12 +1382,12 @@ char *argv[];
   convertpnm (ifp, afp, tfp);
 
   if (alpha)
-    pm_close (afp);
+    pm_closer (afp);
   if ((text) || (ztxt))
-    pm_close (tfp);
+    pm_closer (tfp);
 
-  pm_close (ifp);
-  pm_close (stdout);
+  pm_closer (ifp);
+  pm_closew (stdout);
 
   exit (0);
 }
