@@ -34,6 +34,9 @@
   -Bryan 00.03.05
 */
 
+#define _BSD_SOURCE 1
+    /* Make sure strdup() is in string.h */
+
 #include <string.h>
 #include "pnm.h"
 #ifdef VMS
@@ -42,36 +45,35 @@
 #endif
 #include <tiffioP.h>
 #endif
+#ifdef DJGPP
+#include <tiffioP.h>
+#endif
 #include <tiffio.h>
 
+/* The following are in current tiff.h, but so that we can compile against
+   older tiff libraries, we define them here.
+*/
+
+#ifndef PHOTOMETRIC_LOGL
+#define PHOTOMETRIC_LOGL 32844
+#endif
+#ifndef PHOTOMETRIC_LOGLUV
+#define PHOTOMETRIC_LOGLUV 32845
+#endif
+
 #define MAXCOLORS 1024
 #ifndef PHOTOMETRIC_DEPTH
 #define PHOTOMETRIC_DEPTH 32768
 #endif
 
-/* The following form a cursor into the raster buffer */
-static u_char* inP;
-static int bitsleft;
-
 /* The following describe the structure of the raster buffer */
 static int maxval;
+static unsigned short fillorder;
 static unsigned short bps, spp;
 
 /* The NEXTSAMPLE macro effects the reading of a sample from the raster
    buffer via the above cursor into the variable 'sample'.
 */
-#define NEXTSAMPLE \
-    { \
-    if ( bitsleft == 0 ) \
-	{ \
-	++inP; \
-	bitsleft = 8; \
-	} \
-    bitsleft -= bps; \
-    sample = ( *inP >> bitsleft ) & maxval; \
-    }
-
-
 struct cmdline_info {
     /* All the information the user supplied in the command line,
        in a form easy for the program to use.
@@ -87,27 +89,231 @@ struct cmdline_info {
 static void
 parse_command_line(int argc, char ** argv,
                    struct cmdline_info *cmdline_p);
-static void
-pick_rgba_pixel(xelval * const r_p, xelval * const b_p, xelval * const g_p,
-                xelval * const r_a);
-static void
-pick_cmyk_pixel(xelval * const r_p, xelval * const b_p, xelval * const g_p);
+
+static void read_directory(TIFF * const tif,
+                           unsigned short * const fillorder_p,
+                           unsigned short * const bps_p,
+                           unsigned short * const spp_p,
+                           unsigned short * const photomet_p,
+                           unsigned short * const planarconfig_p,
+                           unsigned int * const cols_p,
+                           unsigned int * const rows_p,
+                           const int headerdump) {
+/*----------------------------------------------------------------------------
+   Read various values of TIFF tags from the TIFF directory, and default 
+   them if not in there.  Exit program with error message if required tags
+   aren't there or values are inconsistent or beyond our capabilities.
+   if 'headerdump' is true, issue informational messages about what we find.
+-----------------------------------------------------------------------------*/
+
+    int rc;
+    unsigned short tiff_bps;
+    unsigned short tiff_spp;
+    unsigned short tiff_fillorder;
+
+    if (headerdump)
+        TIFFPrintDirectory(tif, stderr, TIFFPRINT_NONE);
+
+    rc = TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE, &tiff_bps);
+    *bps_p = (rc == 0) ? 1 : tiff_bps;
+
+    if (*bps_p < 1 || (*bps_p > 8 && *bps_p != 16))
+        pm_error("This program can process Tiff images with only "
+                 "1-8 or 16 bits per sample.  The input Tiff image "
+                 "has %d bits per sample.", bps);
+
+    rc = TIFFGetField(tif, TIFFTAG_FILLORDER, &tiff_fillorder);
+    /* I don't know if it is legal for a Tiff file not to have a 
+       FILLORDER tag, but until I added 16 bit Tiff capability in November
+       2000, we never tried to read the tag.  I don't want to break this 
+       program for older image inputs by erroring out on a missing FILLORDER
+       tag here.  -Bryan
+       */
+    *fillorder_p = (rc == 0) ? FILLORDER_MSB2LSB : tiff_fillorder;
+
+    rc = TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &tiff_spp);
+    *spp_p = (rc == 0) ? 1: tiff_spp;
+
+    rc = TIFFGetField(tif, TIFFTAG_PHOTOMETRIC, photomet_p);
+    if (rc == 0)
+        pm_error("PHOTOMETRIC tag is not in Tiff file.  "
+                 "TIFFGetField() of it failed.\n"
+                 "This means the input is not valid Tiff.");
+
+    if (*spp_p > 1) {
+        rc = TIFFGetField(tif, TIFFTAG_PLANARCONFIG, planarconfig_p);
+        if (rc == 0)
+            pm_error("PLANARCONFIG tag is not in Tiff file, though it "
+                     "has more than one sample per pixel.  "
+                     "TIFFGetField() of it failed.  This means the input "
+                     "is not valid Tiff.");
+    } else {
+        *planarconfig_p = PLANARCONFIG_CONTIG;
+    }
+
+    switch (*spp_p) {
+	case 1:
+	case 3:
+	case 4:
+        break;
+	default:
+        pm_error(
+            "This program can only handle 1-channel gray scale "
+            "or 1- or 3-channel color.  The input Tiff file has %d channels "
+            "(samples per pixel)",
+            *spp_p);
+	}
+
+    switch(*planarconfig_p) {
+	case PLANARCONFIG_CONTIG:
+	    break;
+	case PLANARCONFIG_SEPARATE:
+	    if (*photomet_p != PHOTOMETRIC_RGB && 
+            *photomet_p != PHOTOMETRIC_SEPARATED)
+            pm_error("This program can handle separate planes only "
+                     "with RGB (PHOTOMETRIC tag = %d) or SEPARATED "
+                     "(PHOTOMETRIC tag = %d) data.  The input Tiff file " 
+                     "has PHOTOMETRIC tag = %d.",
+                     PHOTOMETRIC_RGB, PHOTOMETRIC_SEPARATED, &photomet_p);
+	    break;
+	default:
+	    pm_error("Unrecognized PLANARCONFIG tag value in Tiff input: %d.\n",
+                 *planarconfig_p);
+	}
+
+    rc = TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, cols_p);
+    if (rc == 0)
+        pm_error("Input Tiff file is invalid.  It has no IMAGEWIDTH tag.");
+    rc = TIFFGetField( tif, TIFFTAG_IMAGELENGTH, rows_p );
+    if (rc == 0)
+        pm_error("Input Tiff file is invalid.  It has no IMAGELENGTH tag.");
+
+    if (headerdump) {
+        pm_message( "%dx%dx%d image", *cols_p, *rows_p, *bps_p * *spp_p );
+        pm_message( "%d bits/sample, %d samples/pixel", *bps_p, *spp_p );
+	}
+}
 
 
 
+static void
+readscanline(TIFF * const tif, unsigned short * const samplebuf,
+             unsigned char scanbuf[], const int row, const int plane,
+             const unsigned int cols, const unsigned short bps, 
+             const unsigned short fillorder) {
+/*----------------------------------------------------------------------------
+   Read one scanline out of the Tiff input and store it into samplebuf[].
+   Unlike the scanline returned by the Tiff library function, samplebuf[]
+   is composed of one sample per array element, which makes it easier for
+   our caller to process.
+
+   scanbuf[] is a scratch array for our use, which is big enough to hold
+   a Tiff scanline.
+-----------------------------------------------------------------------------*/
+    int rc;
+    const unsigned int bpsmask = (1 << bps) - 1;
+      /* A mask for taking the lowest 'bps' bits of a number */
+
+    rc = TIFFReadScanline(tif, scanbuf, row, plane);
+    if (rc < 0)
+        pm_error( "Unable to read row %d, plane %d of input Tiff image.  "
+                  "TIFFReadScanline() failed.",
+                  row, plane);
+    else if (bps == 8) {
+        int sample;
+        for (sample = 0; sample < cols*spp; sample++) 
+            samplebuf[sample] = scanbuf[sample];
+    } else if (bps < 8) {
+        /* Note that in this format, samples do not span bytes.  Rather,
+           each byte may have don't-care bits in the low order positions.
+           At least that's how I infer the format from reading pnmtotiff.c
+           -Bryan 00.11.18
+           */
+        int sample;
+        int bitsleft;
+        unsigned char * inP;
+
+        bitsleft = 0;
+        inP = scanbuf;
+        for (sample = 0; sample < cols*spp; sample++) {
+            if (bitsleft == 0) {
+                ++inP; 
+                bitsleft = 8;
+            } 
+            bitsleft -= bps; 
+            samplebuf[sample] = (*inP >> bitsleft) & bpsmask; 
+        }
+    } else if (bps == 16) {
+        /* Here's an idea to make this faster:  determine the endianness of
+           the machine we're running on, and if it's the same as that of the
+           input ('fillorder'), just do a memcpy().
+           */
+        switch (fillorder) {
+        case FILLORDER_MSB2LSB: {
+            int sample;
+            unsigned char * inP;
+            inP = scanbuf;  /* initial value */
+            for (sample = 0; sample < cols*spp; sample++) {
+                samplebuf[sample] = (*inP << 8) | *(inP+1);
+                inP += 2;
+            }
+        }
+        break;
+        case FILLORDER_LSB2MSB: {
+            int sample;
+            unsigned char * inP;
+            inP = scanbuf;  /* initial value */
+            for (sample = 0; sample < cols*spp; sample++) {
+                samplebuf[sample] = (*(inP+1) << 8) | *inP;
+                inP += 2;
+            }
+        }
+        break;
+        default:
+            pm_error("Internal error: invalid fillorder passed to "
+                     "readscanline()");
+        }
+    } else
+        pm_error("Internal error: invalid bits per sample passed to "
+                 "readscanline()");
+}
+
+
+
+static void
+pick_cmyk_pixel(const unsigned short samplebuf[], const int sample_cursor,
+                xelval * const r_p, xelval * const b_p, xelval * const g_p) {
+
+    unsigned char c, y, m, k;
+
+    c = samplebuf[sample_cursor+0];
+    y = samplebuf[sample_cursor+1];
+    m = samplebuf[sample_cursor+2];
+    k = samplebuf[sample_cursor+3];
+
+    /* I don't know why this works.  It seems to me that yellow pigment should
+       translate into both red and green light, not just green as you see here.
+       But this is the formula used by TIFFRGBAImageGet() in the Tiff library.
+       */
+    
+    *r_p = ((255-k)*(255-c))/255;
+    *g_p = ((255-k)*(255-y))/255;
+    *b_p = ((255-k)*(255-m))/255;
+}
+
 
 int
-main( argc, argv )
-    int argc;
-    char* argv[];
-    {
-    int cols, rows, grayscale, format;
+main(int argc, char * argv[]) {
+
+    unsigned int cols, rows;
+    int grayscale, format;
     int numcolors;
     register TIFF* tif;
     FILE *alpha_file, *imageout_file;
     int row, i;
     register int col;
-    u_char* buf;
+    unsigned char * scanbuf;
+    unsigned short * samplebuf;
     xel* xelrow;
         /* The ppm-format row of the image row we are presently converting */
     gray* alpharow;
@@ -115,7 +321,7 @@ main( argc, argv )
            row we are presently converting.
            */
     xel colormap[MAXCOLORS];
-    register u_char sample;
+    unsigned char sample;
     unsigned short photomet, planarconfig, inkset;
     unsigned short* redcolormap;
     unsigned short* greencolormap;
@@ -148,63 +354,15 @@ main( argc, argv )
     else
         imageout_file = stdout;
 
+    read_directory(tif, &fillorder, &bps, &spp, &photomet, &planarconfig, 
+                   &cols, &rows, 
+                   cmdline.headerdump);
 
-    if ( cmdline.headerdump )
-	TIFFPrintDirectory( tif, stderr, TIFFPRINT_NONE );
-
-    if ( ! TIFFGetField( tif, TIFFTAG_BITSPERSAMPLE, &bps ) )
-	bps = 1;
-    if ( ! TIFFGetField( tif, TIFFTAG_SAMPLESPERPIXEL, &spp ) )
-	spp = 1;
-    if ( ! TIFFGetField( tif, TIFFTAG_PHOTOMETRIC, &photomet ) )
-	pm_error( "error getting photometric" );
-    if( spp > 1 ){
-	if ( ! TIFFGetField( tif, TIFFTAG_PLANARCONFIG, &planarconfig ) )
-	    pm_error( "error getting planarconfig" );
-    }else{
-	planarconfig = PLANARCONFIG_CONTIG;
-    }
-
-
-    switch ( spp )
-	{
-	case 1:
-	case 3:
-	case 4:
-	break;
-
-	default:
-	pm_error(
-	    "can only handle 1-channel gray scale or 1- or 3-channel color" );
-	}
-
-    switch( planarconfig )
-	{
-	case PLANARCONFIG_CONTIG:
-	    break;
-	case PLANARCONFIG_SEPARATE:
-	    if (photomet != PHOTOMETRIC_RGB && photomet != PHOTOMETRIC_SEPARATED)
-		pm_error( "can only handle separate planes "
-                 "with RGB or SEPARATED data" );
-	    break;
-	default:
-	    pm_error("Unrecongnized PLANARCONFIG tag!\n");
-	}
-
-    (void) TIFFGetField( tif, TIFFTAG_IMAGEWIDTH, &cols );
-    (void) TIFFGetField( tif, TIFFTAG_IMAGELENGTH, &rows );
-
-    if ( cmdline.headerdump )
-	{
-	pm_message( "%dx%dx%d image", cols, rows, bps * spp );
-	pm_message( "%d bits/sample, %d samples/pixel", bps, spp );
-	}
-
-    maxval = ( 1 << bps ) - 1;
-    if ( maxval == 1 && spp == 1 ) {
+    if ( bps == 1 && spp == 1 ) {
         if ( cmdline.headerdump )
-          pm_message("monochrome" );
+            pm_message("monochrome" );
         grayscale = 1;
+        maxval = 1;
 	} else {
         /* How come we don't deal with the photometric for the monochrome 
            case (make sure it's one we know)?  -Bryan 00.03.04
@@ -214,12 +372,14 @@ main( argc, argv )
             if ( cmdline.headerdump )
               pm_message( "%d graylevels (min=black)", maxval + 1 );
             grayscale = 1;
+            maxval = (1 << bps) - 1;
             break;
 
           case PHOTOMETRIC_MINISWHITE:
             if ( cmdline.headerdump )
               pm_message( "%d graylevels (min=white)", maxval + 1 );
             grayscale = 1;
+            maxval = (1 << bps) - 1;
             break;
 
           case PHOTOMETRIC_PALETTE:
@@ -228,9 +388,9 @@ main( argc, argv )
             if ( ! TIFFGetField( tif, TIFFTAG_COLORMAP, 
                                 &redcolormap, &greencolormap, &bluecolormap ) )
               pm_error( "error getting colormaps" );
-            numcolors = maxval + 1;
+            numcolors = 1 << bps;
             if ( numcolors > MAXCOLORS )
-              pm_error( "too many colors" );
+                pm_error( "too many colors" );
             maxval = PNM_MAXMAXVAL;
             grayscale = 0;
             for ( i = 0; i < numcolors; ++i ) {
@@ -255,12 +415,14 @@ main( argc, argv )
                        "pixel.  We need 4 samples, though: C, M, Y, and K.  ",
                        spp);
             grayscale = 0;
+            maxval = (1 << bps) - 1;
             break;
             
           case PHOTOMETRIC_RGB:
             if ( cmdline.headerdump )
               pm_message( "RGB truecolor" );
             grayscale = 0;
+            maxval = (1 << bps) - 1;
             break;
 
           case PHOTOMETRIC_MASK:
@@ -285,11 +447,9 @@ main( argc, argv )
             pm_error( "unknown photometric: %d", photomet );
 	    }
 	}
-    if ( maxval > PNM_MAXMAXVAL )
-	pm_error("bits/sample is too large - "
-             "try reconfiguring Tifftopnm with PGM_BIGGRAYS\n"
-             "    or without PPM_PACKCOLORS" );
-
+    if ( maxval > PNM_OVERALLMAXVAL )
+        pm_error("bits/sample (%d) in the input image is too large.\n",
+                 bps);
     if ( grayscale )
 	{
 	if ( maxval == 1 )
@@ -309,9 +469,13 @@ main( argc, argv )
 	pm_message( "writing PPM file" );
 	}
 
-    buf = (u_char*) malloc(TIFFScanlineSize(tif));
-    if ( buf == NULL )
-	pm_error( "can't allocate memory for scanline buffer" );
+    scanbuf = (unsigned char *) malloc(TIFFScanlineSize(tif));
+    if (scanbuf == NULL)
+        pm_error("can't allocate memory for scanline buffer");
+
+    samplebuf = (unsigned short *) malloc(cols * sizeof(unsigned short) * spp);
+    if (samplebuf == NULL)
+        pm_error ("can't allocate memory for row buffer");
 
     if (imageout_file != NULL) 
         pnm_writepnminit( imageout_file, 
@@ -323,44 +487,44 @@ main( argc, argv )
     alpharow = pgm_allocrow( cols );
 
     for ( row = 0; row < rows; ++row ) {
-        if ( TIFFReadScanline( tif, buf, row, 0 ) < 0 )
-          pm_error( "bad data read on line %d", row );
-
-        /* Set readout cursor to beginning of raster buffer */
-        inP = buf;
-        bitsleft = 8;
+        /* Read one row of samples into samplebuf[] */
+        readscanline(tif, samplebuf, scanbuf, row, 0, 
+                     cols, bps, fillorder);
 
         switch ( photomet ) {
           case PHOTOMETRIC_MINISBLACK:
             for ( col = 0; col < cols; ++col ) {
-                NEXTSAMPLE
-                PNM_ASSIGN1( xelrow[col], sample );
+                PNM_ASSIGN1( xelrow[col], samplebuf[col] );
                 alpharow[col] = 0;
             }
             break;
 
           case PHOTOMETRIC_MINISWHITE:
             for ( col = 0; col < cols; ++col ) {
-                NEXTSAMPLE
                 sample = maxval - sample;
-                PNM_ASSIGN1( xelrow[col], sample );
+                PNM_ASSIGN1( xelrow[col], samplebuf[col] );
                 alpharow[col] = 0;
             }
             break;
 
           case PHOTOMETRIC_PALETTE:
             for ( col = 0; col < cols; ++col ) {
-                NEXTSAMPLE
-                xelrow[col] = colormap[sample];
+                /* We know the following array index is in bounds because
+                   we filled samplebuf with samples of 'bps' bits each and
+                   we verified that the largest number that fits in 'bps'
+                   bits is less than MAXCOLORS, the dimension of the array.
+                   */
+                xelrow[col] = colormap[samplebuf[col]];
                 alpharow[col] = 0;
             }
             break;
 
           case PHOTOMETRIC_SEPARATED:
             if (planarconfig == PLANARCONFIG_CONTIG) {
-                for ( col = 0; col < cols; ++col ) {
+                int col, sample;
+                for ( col = 0, sample = 0; col < cols; ++col, sample+=spp) {
                     xelval r, g, b;
-                    pick_cmyk_pixel(&r, &b, &g);
+                    pick_cmyk_pixel( samplebuf, sample, &r, &b, &g );
             
                     PPM_ASSIGN( xelrow[col], r, g, b );
                     alpharow[col] = 0;
@@ -371,46 +535,39 @@ main( argc, argv )
 
           case PHOTOMETRIC_RGB:
             if( planarconfig == PLANARCONFIG_CONTIG ) {
-                for ( col = 0; col < cols; ++col ) {
-                    xelval r, g, b, a;
-                    
-                    pick_rgba_pixel(&r, &b, &g, &a);
-            
-                    PPM_ASSIGN( xelrow[col], r, g, b );
-                    alpharow[col] = a;
+                int col, sample;
+                for ( col = 0, sample = 0; col < cols; ++col, sample+=spp ) {
+                    PPM_ASSIGN( xelrow[col], samplebuf[sample+0],
+                                samplebuf[sample+1], samplebuf[sample+2] );
+                    if (spp < 4)
+                        alpharow[col] = samplebuf[sample+3];
+                    else
+                        alpharow[col] = 0;
                 }
             } else {
                 /* The input is in separate planes, so the scan line in the 
                    buffer now is just the red.  We need to read twice more
                    to get the green and blue.
                    */
+                int col;
                 
                 /* First clear the value and assign the reds */
                 for ( col = 0; col < cols; ++col ) {
                     PPM_ASSIGN( xelrow[col], 0, 0, 0 );
-                    NEXTSAMPLE
-                    PPM_PUTR( xelrow[col], sample );
+                    PPM_PUTR( xelrow[col], samplebuf[col] );
                 }
                 
                 /* Next the greens */
-                if ( TIFFReadScanline( tif, buf, row, 1 ) < 0 )
-                  pm_error( "bad data read on green line %d", row );
-                inP = buf;
-                bitsleft = 8;
-                for ( col = 0; col < cols; ++col ) {
-                    NEXTSAMPLE
-                    PPM_PUTG( xelrow[col], sample );
-                }
+                readscanline( tif, samplebuf, scanbuf, row, 1,
+                             cols, bps, fillorder );
+                for ( col = 0; col < cols; ++col ) 
+                    PPM_PUTG( xelrow[col], samplebuf[col] );
 
                 /* And finally the blues */
-                if ( TIFFReadScanline( tif, buf, row, 2 ) < 0 )
-                  pm_error( "bad data read on green line %d", row );
-                inP = buf;
-                bitsleft = 8;
-                for ( col = 0; col < cols; ++col ) {
-                    NEXTSAMPLE
-                    PPM_PUTB( xelrow[col], sample );
-                }
+                readscanline( tif, samplebuf, scanbuf, row, 2,
+                             cols, bps, fillorder );
+                for ( col = 0; col < cols; ++col ) 
+                    PPM_PUTB( xelrow[col], samplebuf[col] );
                 /* Could there be an alpha plane?  (We assume no.  But if so,
                    here is where to read it) */
                 for ( col = 0; col < cols; ++col) 
@@ -442,62 +599,6 @@ main( argc, argv )
 
 
 
-static void
-pick_rgba_pixel(xelval * const r_p, xelval * const b_p, xelval * const g_p,
-                xelval * const a_p) {
-/*----------------------------------------------------------------------------
-   Pull the next pixel from the TIFF input, assuming it is in rgb or
-   rgba format (as determined by the global value 'spp', the number of
-   samples per pixel).
-
-   Return the individual components of the pixel.  If there is no alpha
-   component, return 0 as the alpha component.
------------------------------------------------------------------------------*/
-
-    register u_char sample;
-
-    NEXTSAMPLE
-    *r_p = sample;
-    NEXTSAMPLE
-    *g_p = sample;
-    NEXTSAMPLE
-    *b_p = sample;
-    if ( spp == 4 ) {
-        NEXTSAMPLE 
-        *a_p = sample;
-    } else 
-        *a_p = 0;
-}
-
-
-
-static void
-pick_cmyk_pixel(xelval * const r_p, xelval * const b_p, xelval * const g_p) {
-
-    register u_char sample;
-    register u_char c, y, m, k;
-    
-
-    NEXTSAMPLE
-    c = sample;
-    NEXTSAMPLE
-    y = sample;
-    NEXTSAMPLE
-    m = sample;
-    NEXTSAMPLE
-    k = sample;
-
-    /* I don't know why this works.  It seems to me that yellow pigment should
-       translate into both red and green light, not just green as you see here.
-       But this is the formula used by TIFFRGBAImageGet() in the Tiff library.
-       */
-    
-    *r_p = ((255-k)*(255-c))/255;
-    *g_p = ((255-k)*(255-y))/255;
-    *b_p = ((255-k)*(255-m))/255;
-}
-
-
 static void
 parse_command_line(int argc, char ** argv,
                    struct cmdline_info *cmdline_p) {
