@@ -14,7 +14,6 @@
 #include "ppm.h"
 
 static void canonstr ARGS((char* str));
-static long rgbnorm ARGS((long rgb, long lmaxval, int n, char* colorname));
 static void
 canonstr( str )
     char* str;
@@ -32,35 +31,112 @@ canonstr( str )
         }
     }
 
-static long
-rgbnorm( rgb, lmaxval, n, colorname )
-    long rgb, lmaxval;
-    int n;
-    char* colorname;
-    {
-    switch ( n )
-        {
-        case 1:
-        if ( lmaxval != 15 )
-            rgb = rgb * lmaxval / 15;
-        break;
-        case 2:
-        if ( lmaxval != 255 )
-            rgb = rgb * lmaxval / 255;
-        break;
-        case 3:
-        if ( lmaxval != 4095 )
-            rgb = rgb * lmaxval / 4095;
-        break;
-        case 4:
-        if ( lmaxval != 65535L )
-            rgb = rgb * lmaxval / 65535L;
-        break;
-        default:
-        pm_error( "invalid color specifier - \"%s\"", colorname );
+static FILE *
+open_colorname_file(const int must_open) {
+/*----------------------------------------------------------------------------
+   Open the colorname database file.  Its pathname is the value of the
+   environment variable whose name is RGB_ENV (e.g. "RGBDEF").  Except
+   if that environment variable is not set, it is RGB_DB1, RGB_DB2,
+   or RGB_DB3 (e.g. "/usr/lib/X11/rgb.txt"), whichever exists.
+   
+   'must_open' is a logical: we must get the file open or die.  If
+   'must_open' is true and we can't open the file (e.g. it doesn't
+   exist), exit the program with an error message.  If 'must_open' is
+   false and we can't open the file, just return a null pointer.
+-----------------------------------------------------------------------------*/
+    const char *rgbdef;
+    FILE *f;
+
+    if ((rgbdef = getenv(RGBENV))==NULL) {
+        /* The environment variable isn't set, so try the hardcoded
+           default color names database locations.
+        */
+        if ((f = fopen(RGB_DB1, "r")) == NULL &&
+            (f = fopen(RGB_DB2, "r")) == NULL &&
+            (f = fopen( RGB_DB3, "r")) == NULL && must_open) {
+            pm_error("can't open color names database file named "
+                     "%s, %s, or %s "
+                     "and Environment variable %s not set.  Set %s to "
+                     "the pathname of your rgb.txt file or don't use "
+                     "color names.", 
+                     RGB_DB1, RGB_DB2, RGB_DB3, RGBENV, RGBENV);
         }
-    return rgb;
+    } else {            
+        /* The environment variable is set */
+        if ((f = fopen(rgbdef, "r")) == NULL && must_open)
+            pm_error("Can't open the color names database file named %s, "
+                     "per the %s environment variable.",
+                     rgbdef, RGBENV);
     }
+    return(f);
+}
+
+
+static int
+colorget(long * const r_p, long * const g_p, long * const b_p, 
+         char ** const colorname_p, FILE *f) {
+/*----------------------------------------------------------------------------
+   Get next color entry from the color name database file 'f'.a
+
+   If eof or error, return FALSE;  Else, return TRUE.
+
+   return color name in static storage within.
+-----------------------------------------------------------------------------*/
+    char buf[200];
+    static char colorname[200];
+    int got_one;
+    
+    got_one = FALSE;  /* initial value */
+    while (!got_one && fgets(buf, sizeof(buf), f ) != NULL) {
+        if (buf[0] != '\0' && buf[0] != '!' && buf[0] != '#') {
+            if (sscanf(buf, "%ld %ld %ld %[^\n]", r_p, g_p, b_p, colorname) 
+                != 4 ) 
+                pm_message("can't parse color names database line - \"%s\"", 
+                           buf);
+            else got_one = TRUE;
+        }
+    }
+    *colorname_p = colorname;
+    return (got_one);
+}
+
+
+static long
+rgbnorm(const long rgb, const long lmaxval, const int n, 
+        const char * const colorname) {
+/*----------------------------------------------------------------------------
+   Normalize the color (r, g, or b) value 'rgb', which was specified with
+   'n' digits, to a maxval of 'lmaxval'.  If the number of digits isn't
+   valid, issue an error message and identify the complete color 
+   color specification in error as 'colorname'.
+
+   For example, if the user says 0ff/000/000 and the maxval is 100,
+   then rgb is 0xff, n is 3, and our result is 
+   0xff / (16**3-1) * 100 = 6.
+
+-----------------------------------------------------------------------------*/
+    long retval;
+
+    switch (n) {
+    case 1:
+        retval = (long)((double) rgb * lmaxval / 15 + 0.5);
+        break;
+    case 2:
+        retval = (long) ((double) rgb * lmaxval / 255 + 0.5);
+        break;
+    case 3:
+        retval = (long) ((double) rgb * lmaxval / 4095 + 0.5);
+        break;
+    case 4:
+        retval = (long) ((double) rgb * lmaxval / 65535L + 0.5);
+        break;
+    default:
+        pm_error( "invalid color specifier - \"%s\"", colorname );
+    }
+    return retval;
+}
+
+
 
 #if __STDC__
 pixel
@@ -205,43 +281,23 @@ ppm_parsecolor( colorname, maxval )
     else
         {
         /* Must be a name from the X-style rgb file. */
-#ifndef RGB_DB
-        pm_error( "color names database required - please reconfigure with RGBDEF" );
-#else /*RGB_DB*/
         FILE* f;
-        char buf1[200], buf2[200];
+        char *this_colorname;
+        int gotit;
 
-#ifndef A_RGBENV
-        (void) sprintf( buf1, "%s.txt", RGB_DB );
-        if ( ( f = fopen( buf1, "r" ) ) == NULL )
-            pm_error( "can't open color names database - reconfigure with correct RGBDEF" );
-#else /* A_RGBENV */
-        char *rgbdef;
-        if( (rgbdef = getenv(RGB_DB))==NULL )
-            pm_error( "can't get path to color names database - %s not set", RGB_DB );
-        if ( ( f = fopen( rgbdef, "r" ) ) == NULL )
-            pm_error( "can't open color names database - set %s to correct path", RGB_DB );
-#endif /* A_RGBENV */
+        f = open_colorname_file(TRUE);  /* exits if error */
         canonstr( colorname );
-        while ( fgets( buf1, sizeof(buf1), f ) != NULL ) {
-            if (buf1[0] != '\0' && buf1[0] != '!' && buf1[0] != '#') {
-                if ( sscanf( buf1, "%ld %ld %ld %[^\n]", &r, &g, &b, buf2 ) 
-                     != 4 ) {
-                    pm_message(
-                        "can't parse color names database line - \"%s\"", 
-                        buf1 );
-                    continue;
-                }
-                canonstr( buf2 );
-                if ( strcmp( colorname, buf2 ) == 0 )
-                    goto gotit;
-            }
+        gotit = FALSE;
+        while ( !gotit && colorget( &r, &g, &b, &this_colorname, f ) ) {
+            canonstr( this_colorname );
+            if ( strcmp( colorname, this_colorname ) == 0 )
+                gotit = TRUE;
         }
         (void) fclose( f );
-        pm_error( "unknown color - \"%s\"", colorname );
 
-gotit:
-        (void) fclose( f );
+        if (!gotit)
+            pm_error( "unknown color - \"%s\"", colorname );
+
         /* Rescale from [0..255] if necessary. */
         if ( lmaxval != 255 )
             {
@@ -249,7 +305,6 @@ gotit:
             g = g * lmaxval / 255;
             b = b * lmaxval / 255;
             }
-#endif /*RGB_DB*/
         }
 
     PPM_ASSIGN( p, r, g, b );
@@ -270,16 +325,7 @@ ppm_colorname( colorP, maxval, hexok )
 #endif /*__STDC__*/
     {
     int r, g, b;
-#ifdef RGB_DB
     FILE* f;
-    char buf[200];
-    int this_r, this_g, this_b;
-    int best_diff, this_diff;
-    char this_colorname[200];
-#ifdef A_RGBENV
-    char *rgbdef;
-#endif /* A_RGBENV */
-#endif /*RGB_DB*/
 
     if ( maxval == 255 )
         {
@@ -294,44 +340,35 @@ ppm_colorname( colorP, maxval, hexok )
         b = (int) PPM_GETB( *colorP ) * 255 / (int) maxval;
         }
 
-#ifdef RGB_DB
-#ifndef A_RGBENV
-    (void) sprintf( buf, "%s.txt", RGB_DB );
-    if ( ( f = fopen( buf, "r" ) ) == NULL )
-        pm_error( "can't open color names database - reconfigure with correct RGBDEF" );
-#else /* A_RGBENV */
-    if( (rgbdef = getenv(RGB_DB))==NULL )
-        pm_error( "can't get path to color names database - %s not set", RGB_DB );
-    if ( ( f = fopen( rgbdef, "r" ) ) == NULL )
-        pm_error( "can't open color names database - set %s to correct path", RGB_DB );
-#endif /* A_RGBENV */
-    best_diff = 32767;
-    while ( fgets( buf, sizeof(buf), f ) != NULL )
+    f = open_colorname_file(!hexok);
+    if (f != NULL) {
+        long this_r, this_g, this_b;
+        char *this_colorname;
+        int best_diff, this_diff;
+
+        best_diff = 32767;
+        while (colorget(&this_r, &this_g, &this_b, &this_colorname, f)) 
         {
-        if ( sscanf( buf, "%d %d %d %[^\n]", &this_r, &this_g, &this_b,
-                     this_colorname ) != 4 )
+            this_diff = 
+                abs( r - this_r ) + abs( g - this_g ) + abs( b - this_b );
+            if ( this_diff < best_diff )
             {
-            pm_message(
-                "can't parse color names database line - \"%s\"",
-                buf );
-            continue;
-            }
-        this_diff = abs( r - this_r ) + abs( g - this_g ) + abs( b - this_b );
-        if ( this_diff < best_diff )
-            {
-            best_diff = this_diff;
-            (void) strcpy( colorname, this_colorname );
+                best_diff = this_diff;
+                (void) strcpy( colorname, this_colorname );
             }
         }
-    (void) fclose( f );
-    if ( best_diff != 32767 && ( best_diff == 0 || ! hexok ) )
-        return colorname;
-#endif /*RGB_DB*/
+        (void) fclose( f );
+        if ( best_diff != 32767 && ( best_diff == 0 || ! hexok ) )
+            return colorname;
+    }
 
-    /* Color lookup failed; generate an X11-style hex specifier. */
-    if ( ! hexok )
-        pm_error(
-            "color names database required - please reconfigure with RGBDEF" );
+    /* Color lookup failed, but caller is willing to take an X11-style
+       hex specifier, so return that.
+    */
     sprintf( colorname, "#%02x%02x%02x", r, g, b );
     return colorname;
     }
+
+
+
+
