@@ -15,14 +15,14 @@
   Copyright (C) 1991-1998, Thomas G. Lane.
 
   TODO:
-    - Make in pnmtojpeg instead of ppmtojpeg.  If input is PGM or PBM,
+    - Make it pnmtojpeg instead of ppmtojpeg.  If input is PGM or PBM,
       produce JCS_GRAYSCALE output directly.
 
-    - handle 16 bit pixval.  Need to document this an official PGM and PPM
-      format too.
-
 *****************************************************************************/
 
+#define _BSD_SOURCE 1
+    /* This makes sure strdup() is in string.h */
+
 #include <ctype.h>		/* to declare isdigit(), etc. */
 #include <stdlib.h>
 #include <string.h>
@@ -61,6 +61,7 @@ struct cmdline_info {
   unsigned int restart_value;
   enum restart_unit restart_unit;
   char *restart;
+  char *comment;
 } cmdline;
 
 static void
@@ -87,7 +88,7 @@ static boolean set_sample_factors (j_compress_ptr cinfo, char *arg);
 
 
 int
-main(const int argc, char ** argv) {
+main(int argc, char ** argv) {
   struct jpeg_compress_struct cinfo;
   struct jpeg_error_mgr jerr;
   FILE *input_file;
@@ -108,6 +109,8 @@ main(const int argc, char ** argv) {
   int quality;
   int q_scale_factor;
 
+  ppm_init(&argc, argv);
+
   progname = argv[0];
 
   parse_command_line(argc, argv, &cmdline);
@@ -129,7 +132,7 @@ main(const int argc, char ** argv) {
   if (cmdline.input_filename == NULL) 
     input_file = stdin;
   else {
-    if ((input_file = fopen(cmdline.input_filename, "r")) == NULL)
+    if ((input_file = fopen(cmdline.input_filename, "rb")) == NULL)
       pm_error("can't open %s.  Errno=%s(%d).\n", 
                cmdline.input_filename, strerror(errno), errno);
     free(cmdline.input_filename);
@@ -217,6 +220,10 @@ main(const int argc, char ** argv) {
   /* Start compressor */
   jpeg_start_compress(&cinfo, TRUE);
 
+  if (cmdline.comment) 
+      jpeg_write_marker(&cinfo, JPEG_COM, (const JOCTET *) cmdline.comment, 
+                        strlen(cmdline.comment));
+
   /* Translate and copy over the actual scanlines */
   convert_scanlines(&cinfo, input_file, maxval, input_fmt, rescale);
 
@@ -319,6 +326,8 @@ parse_command_line(const int argc, char ** argv,
     optStruct *option_def = malloc(100*sizeof(optStruct));
     /* Instructions to OptParseOptions on how to parse our options.
      */
+    optStruct2 opt;
+
     unsigned int option_def_index;
 
     int argc_parse;       /* argc, except we modify it as we parse */
@@ -344,6 +353,7 @@ parse_command_line(const int argc, char ** argv,
     OPTENTRY(0, "optimize",    OPT_FLAG,   &cmdline_p->optimize,       0);
     OPTENTRY(0, "optimise",    OPT_FLAG,   &cmdline_p->optimize,       0);
     OPTENTRY(0, "restart",     OPT_STRING, &restart,                   0);
+    OPTENTRY(0, "comment",     OPT_STRING, &cmdline_p->comment,        0);
 
     /* Set the defaults */
     cmdline_p->verbose = FALSE;
@@ -362,13 +372,18 @@ parse_command_line(const int argc, char ** argv,
     cmdline_p->smoothing_factor = -1;
     cmdline_p->optimize = FALSE;
     restart = NULL;
+    cmdline_p->comment = NULL;
 
     /* Make private copy of arguments for optParseOptions to corrupt */
     argc_parse = argc;
     for (i=0; i < argc; i++) argv_parse[i] = argv[i];
 
-    pm_optParseOptions(&argc_parse, argv_parse, option_def, 0);
-    /* Uses and sets argc_parse, argv_parse and all of *cmdline_p. */
+    opt.opt_table = option_def;
+    opt.short_allowed = FALSE;  /* We have no short (old-fashioned) options */
+    opt.allowNegNum = FALSE;  /* We have no parms that are negative numbers */
+
+    pm_optParseOptions2(&argc_parse, argv, opt, 0);
+        /* Uses and sets argc_parse, argv, and some of *cmdline_p and others. */
 
     if (argc_parse - 1 == 0)
         cmdline_p->input_filename = NULL;  /* he wants stdin */
@@ -582,7 +597,7 @@ read_quant_tables (j_compress_ptr cinfo, char * filename,
   long val;
   unsigned int table[DCTSIZE2];
 
-  if ((fp = fopen(filename, "r")) == NULL) {
+  if ((fp = fopen(filename, "rb")) == NULL) {
     fprintf(stderr, "%s: Can't open table file %s\n", progname, filename);
     return FALSE;
   }
